#ifndef CROISE_H_
#define CROISE_H_
#include <eoOp.h>
/**
 * @class eoCPTEO_V1QuadCrossover
 * @author Jacques Bibai
 * @brief definition d'un operateur de croisement  
 */

template<class GenotypeT> 
class eoCPTEO_V1QuadCrossover: public eoQuadOp<GenotypeT>
{
public:
   /**
 * Constructeur vide
 */
  eoCPTEO_V1QuadCrossover()
  
  {
   
  }

 
  string className() const { return "eoCPTEO_V1QuadCrossover"; }

 /**
  * @fn bool operator()(GenotypeT&, GenotypeT &)
  * @brief cet operateur defini comment sera effectue un croisement entre deux individus :
  * nous avons considere ici un croisement par bloc.
  * @param _genotype1 : individu 1
  * @param _genotype2 : individu 2
  * @return booleen : true si l'operation c'est bien deroulee et false sinon 
  */ 
  bool operator()(GenotypeT& _genotype1, GenotypeT & _genotype2) 
  {
    bool oneAtLeastIsModified=true;
    //cout << "*******************Avant le croisement**********************\n";
     //cout << _genotype1;
   //cout << _genotype2;
   if(_genotype1.getGenome().size()==1 && _genotype2.getGenome().size()==1)return false;
    
    unsigned pos1=rng.random(_genotype1.getGenome().size());
    unsigned pos2=rng.random(_genotype2.getGenome().size());
    std::vector<Gene> tmp1;
    tmp1.resize(pos1+_genotype2.getGenome().size()-pos2);
    std::vector<Gene> tmp2;
    tmp2.resize(pos2+_genotype1.getGenome().size()-pos1);
    for(size_t i=0;i<=pos1;i++){
	   		tmp1[i].getGene().resize(_genotype1.get(i).size());
	   		 for(size_t j=0;j<tmp1[i].size();j++){
	   		 	tmp1[i].set(j,_genotype1.get(i).get(j));
	   		 }
	   		_genotype1.get(i).getGene().clear();
	   	}
	   	size_t pos=pos2+1;
	 for(size_t i=pos1+1;i<tmp1.size();i++){
	   		tmp1[i].getGene().resize(_genotype2.get(pos).size());
	   		 for(size_t j=0;j<tmp1[i].size();j++){
	   		 	tmp1[i].set(j,_genotype2.get(pos).get(j));
	   		 }
	   		 _genotype2.get(pos).getGene().clear();
	   		 pos++;
	   	} 
	 
	 for(size_t i=0;i<=pos2;i++){
	   		tmp2[i].getGene().resize(_genotype2.get(i).size());
	   		 for(size_t j=0;j<tmp2[i].size();j++){
	   		 	tmp2[i].set(j,_genotype2.get(i).get(j));
	   		 }
	   		 _genotype2.get(i).getGene().clear();
	   	}
	   	
	   pos=pos1+1;
	 for(size_t i=pos2+1;i<tmp2.size();i++){
	   		tmp2[i].getGene().resize(_genotype1.get(pos).size());
	   		 for(size_t j=0;j<tmp2[i].size();j++){
	   		 	tmp2[i].set(j,_genotype1.get(pos).get(j));
	   		 }
	   		_genotype1.get(pos).getGene().clear();
	   		 pos++;
	   	} 
	   	 	
     _genotype1.getGenome().clear();
     _genotype2.getGenome().clear();
     
      _genotype1.getGenome().resize(tmp2.size());
      _genotype2.getGenome().resize(tmp1.size());
     
      
      for(size_t i =0;i<_genotype1.getGenome().size();i++){
			_genotype1.set(i,tmp2[i]);
		}
		
		for(size_t i =0;i<_genotype2.getGenome().size();i++){
			_genotype2.set(i,tmp1[i]);
		}
      		
    return oneAtLeastIsModified;
    
  }

private:

};
#endif /*CROISE_H_*/
